<?php

namespace Drupal\paragraphs_gantt\Plugin\Field\FieldFormatter;

use Drupal\Component\Serialization\Json;
use Drupal\Component\Utility\Html;
use Drupal\Core\Datetime\DateFormatterInterface;
use Drupal\Core\Entity\EntityDisplayRepositoryInterface;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Entity\EntityReferenceSelection\SelectionPluginManagerInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Field\FieldConfigInterface;
use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\Core\Field\FieldItemListInterface;
use Drupal\Core\Field\Plugin\Field\FieldFormatter\EntityReferenceFormatterBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Path\CurrentPathStack;
use Drupal\Core\Render\Markup;
use Drupal\Core\Template\Attribute;
use Drupal\Core\Url;
use Drupal\field\Entity\FieldConfig;
use Drupal\paragraphs\Entity\Paragraph;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Plugin implementation of the paragraphs gantt formatter.
 *
 * @FieldFormatter(
 *   id = "paragraphs_gantt_formatter",
 *   module = "paragraphs_gantt",
 *   label = @Translation("Paragraphs gantt"),
 *   field_types = {
 *     "entity_reference_revisions"
 *   }
 * )
 */
class ParagraphsGanttFormatter extends EntityReferenceFormatterBase {

  /**
   * The entity display repository.
   *
   * @var \Drupal\Core\Entity\EntityDisplayRepositoryInterface
   */
  protected $entityDisplayRepository;

  /**
   * The fields of paragraphs.
   *
   * @var \Drupal\field\Entity\FieldConfig
   */
  protected $fieldsParagraphs;

  /**
   * The selection plugin manager.
   *
   * @var \Drupal\Core\Entity\EntityReferenceSelection\SelectionPluginManagerInterface
   */
  protected $selectionManager;

  /**
   * The date formatter service.
   *
   * @var \Drupal\Core\Datetime\DateFormatterInterface
   */
  protected $dateFormatter;

  /**
   * The entity field manager service.
   *
   * @var \Drupal\Core\Entity\EntityFieldManagerInterface
   */
  protected $entityFieldManager;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The current path stack.
   *
   * @var \Drupal\Core\Path\CurrentPathStack
   */
  protected $currentPathStack;

  /**
   * The module handler service.
   *
   * @var \Drupal\Core\Extension\ModuleHandlerInterface
   */
  protected $moduleHandler;

  /**
   * Constructs a FormatterBase object.
   *
   * @param string $plugin_id
   *   The plugin_id for the formatter.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\Core\Field\FieldDefinitionInterface $field_definition
   *   The definition of the field to which the formatter is associated.
   * @param array $settings
   *   The formatter settings.
   * @param string $label
   *   The formatter label display setting.
   * @param string $view_mode
   *   The view mode.
   * @param array $third_party_settings
   *   Any third party settings.
   * @param \Drupal\Core\Entity\EntityDisplayRepositoryInterface $entity_display_repository
   *   The entity display repository.
   * @param \Drupal\Core\Datetime\DateFormatterInterface $date_formatter
   *   The date formatter.
   * @param \Drupal\Core\Entity\EntityFieldManagerInterface $entity_field_manager
   *   The entity field manager.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\Core\Path\CurrentPathStack $current_path_stack
   *   The current path stack.
   * @param \Drupal\Core\Extension\ModuleHandlerInterface $module_handler
   *   The module handler.
   * @param \Drupal\Core\Entity\EntityReferenceSelection\SelectionPluginManagerInterface $selection_manager
   *   The entity reference selection plugin  manager.
   */
  public function __construct($plugin_id, $plugin_definition, FieldDefinitionInterface $field_definition, array $settings, $label, $view_mode, array $third_party_settings, EntityDisplayRepositoryInterface $entity_display_repository, DateFormatterInterface $date_formatter, EntityFieldManagerInterface $entity_field_manager, EntityTypeManagerInterface $entity_type_manager, CurrentPathStack $current_path_stack, ModuleHandlerInterface $module_handler, SelectionPluginManagerInterface $selection_manager) {
    parent::__construct($plugin_id, $plugin_definition, $field_definition, $settings, $label, $view_mode, $third_party_settings);
    $this->entityDisplayRepository = $entity_display_repository;
    $this->dateFormatter = $date_formatter;
    $this->entityFieldManager = $entity_field_manager;
    $this->entityTypeManager = $entity_type_manager;
    $this->currentPathStack = $current_path_stack;
    $this->moduleHandler = $module_handler;
    $this->selectionManager = $selection_manager;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $plugin_id,
      $plugin_definition,
      $configuration['field_definition'],
      $configuration['settings'],
      $configuration['label'],
      $configuration['view_mode'],
      $configuration['third_party_settings'],
      $container->get('entity_display.repository'),
      $container->get('date.formatter'),
      $container->get('entity_field.manager'),
      $container->get('entity_type.manager'),
      $container->get('path.current'),
      $container->get('module_handler'),
      $container->get('plugin.manager.entity_reference_selection')
    );
  }

  /**
   * {@inheritdoc}
   */
  public static function defaultSettings() {
    // Implement default settings.
    return [
      'view_mode' => 'default',
      'text' => 'field_gantt_text',
      'type' => 'field_gantt_type',
      'start_date' => 'field_gantt_date',
      'show_end' => FALSE,
      'duration' => 'field_gantt_duration',
      'progress' => 'field_gantt_progress',
      'open' => 'field_gantt_open',
      'parent' => 'field_gantt_parent',
      'links' => 'field_gantt_links',
      'owner' => 'field_gantt_owner',
      'planned_date' => 'field_gantt_planned_date',
      'split_actual_planned' => FALSE,
      'planned_duration' => 'field_gantt_planned_duration',
      'custom_field' => '',
      'add_task' => TRUE,
      'edit_task' => TRUE,
      'native_dialog' => FALSE,
      'work_time' => FALSE,
      'send_email' => FALSE,
      'send_notification' => FALSE,
      'cdn' => TRUE,
    ] + parent::defaultSettings();
  }

  /**
   * {@inheritdoc}
   */
  public function settingsForm(array $form, FormStateInterface $form_state) {
    $settingForm = [
      'view_mode' => [
        '#type' => 'select',
        '#options' => $this->entityDisplayRepository->getViewModeOptions($this->getFieldSetting('target_type')),
        '#title' => $this->t('View mode'),
        '#default_value' => $this->getSetting('view_mode'),
        '#required' => TRUE,
      ],
      'text' => [
        '#title' => $this->t('Title'),
        '#description' => $this->t('Field title'),
        '#type' => 'select',
        '#required' => TRUE,
        '#options' => $this->getConfigurableFields([
          'list_integer',
          'list_string',
          'text',
          'text_long',
          'text_with_summary',
          'string',
          'string_long',
        ]),
        '#empty_option' => $this->t('- None -'),
        '#default_value' => $this->getSetting('text'),
      ],
      'start_date' => [
        '#title' => $this->t('Date'),
        '#description' => $this->t('Field date'),
        '#type' => 'select',
        '#required' => TRUE,
        '#options' => $this->getConfigurableFields(['datetime', 'daterange']),
        '#empty_option' => $this->t('- None -'),
        '#default_value' => $this->getSetting('start_date'),
      ],
      'show_end' => [
        '#type' => 'checkbox',
        '#title' => $this->t('Show end day'),
        '#default_value' => $this->getSetting('show_end'),
        '#description' => $this->t("If check it will add column end date."),
      ],
      'type' => [
        '#title' => $this->t('Type'),
        '#description' => $this->t('Field type'),
        '#type' => 'select',
        '#required' => TRUE,
        '#options' => $this->getConfigurableFields(['list_string', 'string']),
        '#empty_option' => $this->t('- None -'),
        '#default_value' => $this->getSetting('type'),
      ],
      'duration' => [
        '#title' => $this->t('Duration'),
        '#description' => $this->t('Field duration'),
        '#type' => 'select',
        '#empty_option' => $this->t('- None -'),
        '#options' => $this->getConfigurableFields(['integer']),
        '#default_value' => $this->getSetting('duration'),
      ],
      'work_time' => [
        '#type' => 'checkbox',
        '#title' => $this->t('Work day'),
        '#default_value' => $this->getSetting('work_time'),
        '#description' => $this->t("Enables calculating the duration of tasks in working time instead of calendar time."),
      ],
      'progress' => [
        '#title' => $this->t('Progress'),
        '#description' => $this->t('Field progress'),
        '#type' => 'select',
        '#empty_option' => $this->t('- None -'),
        '#options' => $this->getConfigurableFields([
          'float',
          'list_float',
          'decimal',
        ]),
        '#default_value' => $this->getSetting('progress'),
      ],
      'open' => [
        '#title' => $this->t('Open'),
        '#description' => $this->t('Field open'),
        '#type' => 'select',
        '#empty_option' => $this->t('- None -'),
        '#options' => $this->getConfigurableFields(['boolean']),
        '#default_value' => $this->getSetting('open'),
      ],
      'parent' => [
        '#title' => $this->t('Parent'),
        '#description' => $this->t('Field parent'),
        '#type' => 'select',
        '#empty_option' => $this->t('- None -'),
        '#options' => $this->getConfigurableFields([
          'integer',
          'list_integer',
        ]),
        '#default_value' => $this->getSetting('parent'),
      ],
      'links' => [
        '#title' => $this->t('Links type'),
        '#description' => $this->t('Field link'),
        '#type' => 'select',
        '#empty_option' => $this->t('- None -'),
        '#options' => $this->getConfigurableFields([
          'double_field',
        ]),
        '#default_value' => $this->getSetting('links'),
      ],
      'owner' => [
        '#title' => $this->t('Resource'),
        '#description' => $this->t('Resource Assignment'),
        '#type' => 'select',
        '#empty_option' => $this->t('- None -'),
        '#options' => $this->getConfigurableFields([
          'integer',
          'entity_reference',
        ]),
        '#default_value' => $this->getSetting('owner'),
      ],
      'add_task' => [
        '#type' => 'checkbox',
        '#title' => $this->t('Add Task'),
        '#default_value' => $this->getSetting('add_task'),
        '#description' => $this->t("Enable posibility 'Add Task' from Gantt chart."),
      ],
      'edit_task' => [
        '#type' => 'checkbox',
        '#title' => $this->t('Edit Task'),
        '#default_value' => $this->getSetting('edit_task'),
        '#description' => $this->t("Enable possibility 'Edit Task' from Gantt chart."),
      ],
      'send_email' => [
        '#type' => 'checkbox',
        '#title' => $this->t('Send an email after the change of status'),
        '#default_value' => $this->getSetting('send_email'),
      ],
      'cdn' => [
        '#title' => $this->t('Use CDN'),
        '#description' => $this->t('If not you can use custom version in libraries'),
        '#type' => 'checkbox',
        '#default_value' => $this->getSetting('cdn'),
      ],
      'baseline' => [
        '#type' => 'details',
        '#title' => $this->t('Base line (version PRO)'),
        '#description' => $this->t('If you want show double gantt to compare actual vs planned date. This method is available only in Pro versions'),
      ],
      'planned_date' => [
        '#type' => 'select',
        '#title' => $this->t('Planned Date field'),
        '#empty_option' => $this->t('- None -'),
        '#options' => $this->getConfigurableFields(['datetime', 'daterange']),
        '#default_value' => $this->getSetting('planned_date'),
        '#description' => $this->t('Select the field that contains the planned date range in the selected row.'),
      ],
      'split_actual_planned' => [
        '#type' => 'checkbox',
        '#title' => $this->t('Split gantt actual and planned date'),
        '#default_value' => $this->getSetting('split_actual_planned'),
        '#description' => $this->t("Uncheck to show date theory and actual date in the same baseline."),
      ],
      'planned_duration' => [
        '#title' => $this->t('Duration'),
        '#description' => $this->t('Planned duration field'),
        '#type' => 'select',
        '#empty_option' => $this->t('- None -'),
        '#options' => $this->getConfigurableFields(['integer']),
        '#default_value' => $this->getSetting('planned_duration'),
        '#states' => [
          'visible' => [
            ':input[name="style_options[split_actual_planned]"]' => ['checked' => TRUE],
          ],
        ],
      ],
      'custom_field' => [
        '#title' => $this->t('Custom field'),
        '#description' => $this->t('Additional column'),
        '#type' => 'select',
        '#empty_option' => $this->t('- None -'),
        '#options' => $this->getConfigurableFields(['integer']),
        '#default_value' => $this->getSetting('custom_field'),
      ],
    ];

    if ($this->moduleHandler->moduleExists('pwa_firebase')) {
      $settingForm['send_notification'] = [
        '#type' => 'checkbox',
        '#title' => $this->t('Send notification to assign user'),
        '#default_value' => $this->getSetting('send_notification'),
      ];
    }
    if ($this->moduleHandler->moduleExists('paragraphs_table')) {
      $settingForm['native_dialog'] = [
        '#type' => 'checkbox',
        '#title' => $this->t('Custom form'),
        '#default_value' => $this->getSetting('native_dialog'),
        '#description' => $this->t("Use native Drupal form to add/edit."),
      ];
    }
    return $settingForm + parent::settingsForm($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function settingsSummary() {
    if (empty($this->fieldsParagraphs)) {
      $this->setFieldsParagraphs();
    }
    $summary = parent::settingsSummary();
    if (!empty($this->getSetting('text'))) {
      $summary[] = $this->t('Field title: @field', ['@field' => $this->getLabelName($this->getSetting('text'))]);
    }
    if (!empty($this->getSetting('start_date'))) {
      $summary[] = $this->t('Field date: @field', ['@field' => $this->getLabelName($this->getSetting('start_date'))]);
    }
    if (!empty($this->getSetting('duration'))) {
      $summary[] = $this->t('Field duration: @field', ['@field' => $this->getLabelName($this->getSetting('duration'))]);
    }
    if (!empty($this->getSetting('progress'))) {
      $summary[] = $this->t('Field progress: @field', ['@field' => $this->getLabelName($this->getSetting('progress'))]);
    }
    if (!empty($this->getSetting('open'))) {
      $summary[] = $this->t('Field open: @field', ['@field' => $this->getLabelName($this->getSetting('open'))]);
    }
    if (!empty($this->getSetting('parent'))) {
      $summary[] = $this->t('Field parent: @field', ['@field' => $this->getLabelName($this->getSetting('parent'))]);
    }
    if (!empty($this->getSetting('owner'))) {
      $summary[] = $this->t('Field users: @field', ['@field' => $this->getLabelName($this->getSetting('owner'))]);
    }
    return $summary;
  }

  /**
   * Get list of fields.
   */
  protected function getConfigurableFields($type = FALSE) {
    if (empty($this->fieldsParagraphs)) {
      $this->setFieldsParagraphs();
    }
    $listField = [];
    foreach ($this->fieldsParagraphs as $field_name => $field) {
      $listField[$field_name] = $field->getLabel();
      if (!empty($type)) {
        if (!is_array($type) && $type != $field->getType()) {
          unset($listField[$field_name]);
        }
        if (is_array($type) && !in_array($field->getType(), $type)) {
          unset($listField[$field_name]);
        }
      }
    }
    return $listField;
  }

  /**
   * {@inheritdoc}
   */
  protected function setFieldsParagraphs() {
    $targetType = $this->getFieldSetting('target_type');
    $targetBundle = array_key_first($this->fieldDefinition->getSetting("handler_settings")["target_bundles"]);
    /** @var \Drupal\paragraphs\ParagraphInterface $paragraphs_entity */
    $paragraphs_entity = $this->entityTypeManager->getStorage($targetType)
      ->create(['type' => $targetBundle]);
    $field_definitions = $paragraphs_entity->getFieldDefinitions();
    foreach ($field_definitions as $field_name => $field) {
      if ($field instanceof FieldConfigInterface || $field instanceof FieldConfig) {
        $this->fieldsParagraphs[$field_name] = $field;
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  protected function getLabelName($fieldName) {
    if (!empty($this->fieldsParagraphs[$fieldName])) {
      return $this->fieldsParagraphs[$fieldName]->getLabel();
    }
    return FALSE;
  }

  /**
   * {@inheritdoc}
   */
  public function viewElements(FieldItemListInterface $items, $langcode) {

    $setting = $this->getSettings();
    $targetType = $this->getFieldSetting('target_type');
    $field_definition = $items->getFieldDefinition();
    $entity_type_id = $field_definition->get('entity_type');
    $selectionHandler = $this->selectionManager->getSelectionHandler($field_definition);
    $bundles = $selectionHandler->entityTypeBundleInfo->getBundleInfo($targetType);
    $field_name = $field_definition->getName();
    $field_label = $field_definition->getLabel();
    $entity = $items->getEntity();
    $entityId = $entity->id();
    $currentURL = $this->currentPathStack->getPath();
    $destination = ['destination' => $currentURL];
    $view_mode = $this->viewMode == 'full' ? 'default' : $this->viewMode;
    $dialog_width = '80%';
    $handlers = $field_definition->getSetting("handler_settings");
    $handler = array_key_first($handlers['target_bundles']);
    $storage = $this->entityTypeManager->getStorage('paragraphs_type');
    $paragraphs_type = $storage->load($handler);

    // Get list owner.
    $owner = [['key' => 0, 'label' => $this->t('- None -')]];
    if (!empty($setting['owner'])) {
      $paragraph = Paragraph::create(['type' => $handler]);
      $fieldDefinitions = $paragraph->getFieldDefinitions();
      $listOwners = $this->selectionManager
        ->getSelectionHandler($fieldDefinitions[$setting['owner']], $paragraph)
        ->getReferenceableEntities();
      if (!empty($listOwners['user'])) {
        $users = $listOwners['user'];
        if (!empty($users[0])) {
          unset($users[0]);
        }
        $colorIndex = 0;
        foreach ($users as $uid => $userName) {
          $owner[] = [
            'key' => $uid,
            'label' => $userName,
            'textColor' => "#FFF",
            'backgroundColor' => $this->generateColor($colorIndex++),
          ];
        }
      }
    }

    // Custom field.
    if (!empty($setting['custom_field']) && !empty($fieldDefinitions[$setting['custom_field']])) {
      $setting['custom_field_name'] = $fieldDefinitions[$setting['custom_field']]->getLabel();
    }

    // Format date.
    $dateType = 'date';
    $dateFormatStorage = $this->entityTypeManager->getStorage('date_format');
    $drupalShortFormat = $dateFormatStorage->load('short')->getPattern();
    $patternDate = explode(' ', $drupalShortFormat);
    $setting['format_date'] = current($patternDate);
    $paragraph = Paragraph::create(['type' => $paragraphs_type->id()]);
    $fieldDefinitions = $paragraph->getFieldDefinitions();
    if (!empty($fieldDefinitions[$setting['start_date']])) {
      $dateSettings = $fieldDefinitions[$setting['start_date']]->getSettings();
      $dateType = $dateSettings["datetime_type"];
    }
    if ($dateType == 'date') {
      $setting['format_date'] = current(explode(' ', $setting['format_date']));
    }

    $path = '';
    if (method_exists($paragraphs_type, 'getIconUrl')) {
      $path = $paragraphs_type->getIconUrl();
    }
    $title = implode(' ', [
      '<i class="bi bi-calendar-plus"></i>',
      $this->t('Add'),
      $bundles[$handler]['label'],
    ]);
    if (!empty($path)) {
      $title = '<img src="' . $path . '" height="25"/> ' . $title;
      $field_label = '<img src="' . $path . '" height="25"/> ' . $field_label;
    }
    $ganttId = $handler . '_' . $entityId;
    $attr = [
      'class' => ['gantt', $handler],
      'id' => Html::getId($ganttId),
    ];
    $lib = 'paragraphs_gantt/paragraphs-gantt';
    if ($setting['cdn']) {
      $lib .= '.cdn';
    }
    $data = $this->convertGanttData($items, $langcode, $setting);
    $elements = [
      '#type' => 'container',
      '#theme' => 'paragraphs_gantt_wrapper',
      '#field_name' => $field_name,
      '#title' => $field_label,
      '#id' => $ganttId,
      '#id_gantt' => $ganttId,
      '#owner' => $owner,
      '#data' => $data,
      "#paragraph_parent_id" => $entityId,
      "#paragraph_parent_type" => $entity_type_id,
      "#paragraph_parent_field_name" => $field_name,
      '#attributes' => new Attribute($attr),
      '#ajax' => Url::fromRoute('paragraphs.gantt', [
        'paragraph_type' => $handler,
        'entity_type' => $entity_type_id,
        'entity_field' => $field_name,
        'entity_id' => $entityId,
      ]),
      '#add' => [
        '#type' => 'link',
        '#url' => Url::fromRoute('paragraphs.add', [
          'paragraph_type' => $handler,
          'entity_type' => $entity_type_id,
          'entity_field' => $field_name,
          'entity_id' => $entityId,
        ])->setOptions(['query' => $destination]),
        '#title' => Markup::create($title),
        '#attributes' => [
          'class' => ['use-ajax' , 'btn' , 'btn-default'],
          'data-dialog-type' => 'modal',
          'data-dialog-options' => Json::encode(['width' => $dialog_width]),
        ],
      ],
      '#settings' => $setting,
      '#attached' => [
        'drupalSettings' => [
          'date_format' => str_replace(
            ['Y', 'm', 'd', 'H', 'i', 's'],
            ['%Y', '%m', '%d', '%H', '%i', '%s'],
            $setting['format_date']
          ),
          'gantt' => [
            $handler => [
              'id' => 'gantt_' . $ganttId,
              'owner' => $setting['owner'],
              'split_actual_planned' => !empty($setting["split_actual_planned"]),
              'add_task' => !empty($setting["add_task"]),
              'edit_task' => !empty($setting["edit_task"]),
              'native_dialog' => !empty($setting["native_dialog"]),
              'work_time' => !empty($setting['work_time']),
              'has_duration' => !empty($setting['duration']),
              'has_planned' => !empty($data['planned']),
              'show_end' => !empty($setting['show_end']),
              'custom_field' => $setting['custom_field'] ?? FALSE,
              'custom_field_name' => $setting['custom_field_name'] ?? FALSE,
              'current_path' => 'destination=' . $this->currentPathStack->getPath(),
              'ajax' => Url::fromRoute('paragraphs.gantt', [
                'paragraph_type' => $handler,
                'view_mode' => $view_mode,
                'entity_type' => $entity_type_id,
                'entity_field' => $field_name,
                'entity_id' => $entityId,
              ])->toString(),
              'import' => Url::fromRoute('paragraphs.import', [
                'paragraph_type' => $handler,
                'view_mode' => $view_mode,
                'entity_type' => $entity_type_id,
                'entity_field' => $field_name,
                'entity_id' => $entityId,
              ])->toString(),
            ],
          ],
        ],
        'library' => [$lib],
      ],
    ];
    if ($setting['custom_field']) {
      $elements['#attached']['library'][] = 'paragraphs_gantt/paragraphs-gantt.chosen';
    }
    if ($setting['split_actual_planned']) {
      $elements["#dataPlanned"] = !empty($data['planned']) ? ['data' => $data['planned']] : FALSE;
      if (!empty($data['links'])) {
        if (!empty($variables["dataPlanned"])) {
          $elements["#dataPlanned"]['links'] = $data['links'];
        }
      }
    }

    return $elements;
  }

  /**
   * {@inheritdoc}
   */
  protected function convertGanttData(FieldItemListInterface $items, string $langcode, array $setting) {

    $entities = $this->getEntitiesToView($items, $langcode);
    // When a display renderer doesn't exist, fall back to the default.
    $field_definition = $items->getFieldDefinition();
    $handlers = $field_definition->getSetting("handler_settings");
    $targetBundle = array_key_first($handlers['target_bundles']);
    $links = [];
    $outData = [];
    $mappingGantt = [
      'text',
      'type',
      'start_date',
      'duration',
      'progress',
      'open',
      'parent',
      'owner',
      'planned_date',
      'planned_duration',
      'custom_field',
    ];
    $fieldDateType = 'date';
    foreach ($entities as $delta => $entity) {
      $fieldDefinitions = $entity->getFieldDefinitions();
      if (!empty($fieldDefinitions[$setting['start_date']])) {
        $fieldDateType = $fieldDefinitions[$setting['start_date']]->getType();
      }
      if ($targetBundle != $entity->bundle()) {
        continue;
      }

      $data[$delta] = ['id' => $entity->id()];
      $dataPlanned = [];
      foreach ($mappingGantt as $ganttField) {
        if (!empty($field_name = $setting[$ganttField]) && !empty($entity->$field_name)) {
          $value = $entity->get($field_name)->getString();
          if (!empty($value)) {
            $data[$delta][$ganttField] = $value;
          }
          if ($ganttField == 'start_date') {
            $data[$delta][$ganttField] = $this->dateGanttType($entity, $field_name, $setting['format_date'], $fieldDateType, 'value');
          }
          if ($ganttField == 'end_date') {
            $data[$delta][$ganttField] = $this->dateGanttType($entity, $field_name, $setting['format_date'], $fieldDateType, 'end_value');
          }

          // Process Planned Date.
          if ($ganttField == 'planned_date') {
            $data[$delta]['planned_start'] = $this->dateGanttType($entity, $field_name, $setting['format_date'], $fieldDateType, 'value');
            $data[$delta]['planned_end'] = $this->dateGanttType($entity, $field_name, $setting['format_date'], $fieldDateType, 'end_value');
            if (!empty($data[$delta]['planned_start'])) {
              $dataPlanned[$delta] = $data[$delta];
            }
          }

          // Process owner.
          if ($ganttField == 'owner' && !$entity->get($field_name)->isEmpty()) {
            $values = $entity->get($field_name)->getValue();
            $data[$delta][$ganttField] = [];
            foreach ($values as $value) {
              if (!empty($value->entity)) {
                $data[$delta][$ganttField][] = $value->entity->id();
              }
              elseif (is_array($value) && !empty($value['target_id'])) {
                $data[$delta][$ganttField][] = $value['target_id'];
              }
            }
          }

          // Process custom field.
          if ($ganttField == 'custom_field') {
            $data[$delta][$field_name] = $value;
          }
        }
      }
      // Process Links.
      if (!empty($setting['links'])) {
        $valueLinks = $entity->get($setting['links'])->getValue();
        foreach ($valueLinks as $valueLink) {
          $links[] = [
            'id' => implode('-', [$entity->id(), $valueLink['first']]),
            'source' => $entity->id(),
            'target' => (int) $valueLink['first'],
            'type' => (int) $valueLink['second'],
          ];
        }
      }

    }
    // Data planned.
    if (!empty($dataPlanned)) {
      foreach ($dataPlanned as $planned) {
        $planned['start_date'] = $planned['planned_start'];
        $planned['duration'] = $this->getDuration($planned);
        $outData['planned'][] = $planned;
      }
    }
    if (!empty($data)) {
      $outData['data'] = array_values($data);
      if (!empty($links)) {
        $outData['links'] = $links;
      }
    }
    return $outData;
  }

  /**
   * {@inheritDoc}
   */
  protected function getDuration($planned = []) {
    if ($planned['planned_end'] == $planned['planned_start']) {
      return 1;
    }
    if (!empty($planned['planned_date'])) {
      [$start, $end] = explode(', ', $planned['planned_date']);
      return (int) abs(strtotime($end) - strtotime($start)) / 86400;
    }
    else {
      $start = str_replace('/', '-', $planned['planned_start']);
      $end = str_replace('/', '-', $planned['planned_end']);
    }

    return (int) abs(strtotime($end) - strtotime($start)) / 86400;
  }

  /**
   * {@inheritDoc}
   */
  protected function dateGanttType($entity, $field_name, $format, $type = 'daterange', $storageType = 'date') {
    $date = '';
    if (!empty($entity->get($field_name)->date)) {
      $date = $entity->get($field_name)->date->format($format);
    }
    if ($type) {
      $dateValue = $entity->get($field_name)->{$storageType};
      if (!empty($dateValue)) {
        $dateValue = $this->dateFormatter->format(strtotime($dateValue), 'short');
        $date = current(explode(' ', $dateValue));
      }
    }
    return $date;
  }

  /**
   * Change color backgroup foreach user.
   *
   * @param int $index
   *   Index user in array.
   *
   * @return string
   *   Code color.
   */
  private function generateColor(int $index) {
    $listColor = ["#03A9F4", "#f57730", "#e157de", "#78909C", "#8D6E63",
      "#800000", "#008080", "#d2691e", "#ff8c00", "#daa520", "#4b0082",
      "#FF0000", "#800080", "#008000", "#808000", "#000080", "#0000FF",
    ];
    if (empty($listColor[$index])) {
      return $listColor[$index % count($listColor)];
    }
    return $listColor[$index];
  }

}
